import typing

import napari.layers
import numpy as np
from .._utils import general
from napari.utils.notifications import show_info
from qtpy.QtCore import QRegularExpression, Slot
from qtpy.QtGui import QRegularExpressionValidator
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFormLayout,
    QLineEdit,
    QPushButton,
    QWidget,
)

from .._utils import filters

if typing.TYPE_CHECKING:
    import napari


class FilterImageWidget(QWidget):
    def __init__(self, napari_viewer: "napari.Viewer"):
        super().__init__()
        self.napari_viewer = napari_viewer
        self._layer = QComboBox(self)
        self._lp_filter_resolution = QLineEdit("20", self)
        self._lp_filter_resolution.setToolTip(
            "Low-Pass filter value in angstrom. Set to 0 to disable low-pass filtering."
        )
        self._hp_filter_resolution = QLineEdit("0", self)
        self._hp_filter_resolution.setToolTip(
            "High-Pass filter value in angstrom. Set to 0 to disable high-pass filtering."
        )
        self._pixel_size = QLineEdit("-1", self)
        self._hp_filter_resolution.setToolTip("Pixel size of the image.")
        self._filter_2d = QCheckBox(self)
        self._hp_filter_resolution.setToolTip(
            "If the layer is 3D, filter slice-by-slice in z direction rather than the 3D volume."
        )
        self._show_mask = QCheckBox(self)
        self._hp_filter_resolution.setToolTip(
            "Mainly for debugging. Visualizes the filter mask."
        )
        self._run_btn = QPushButton("Run", self)

        float_validator = QRegularExpressionValidator(
            QRegularExpression(r"[0-9]*\.?[0-9]*")
        )
        self._lp_filter_resolution.setValidator(float_validator)
        self._hp_filter_resolution.setValidator(float_validator)
        self._pixel_size.setValidator(float_validator)

        layout = QFormLayout()
        layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.setLayout(layout)
        self.layout().addRow("Image layer:", self._layer)
        self.layout().addRow("LP resolution [A]:", self._lp_filter_resolution)
        self.layout().addRow("HP resolution [A]:", self._hp_filter_resolution)
        self.layout().addRow("Pixel size [A/px]:", self._pixel_size)
        self.layout().addRow("Live filter 2D slices", self._filter_2d)
        self.layout().addRow("Show mask", self._show_mask)
        self.layout().addRow(self._run_btn)

        self._layer.currentTextChanged.connect(self._update_pixel_size)
        self._run_btn.clicked.connect(self._run)
        self.napari_viewer.layers.events.inserted.connect(self._update_combo)
        self.napari_viewer.layers.events.removed.connect(self._update_combo)

        self._update_combo()

    @staticmethod
    def update_links(target_layer: napari.layers.Layer, link_layer_id: int):
        target_layer.metadata.setdefault('linked_image_layers', []).append(link_layer_id)

    @Slot()
    def _run(self):
        for attr in [
            "pixel_size",
            "lp_filter_resolution",
            "hp_filter_resolution",
        ]:
            try:
                if getattr(self, attr) < 0:
                    show_info(f"Error: {attr} needs to be larger than 0")
                    return None
            except ValueError:
                show_info(f"Error: {attr} needs to be provided")
                return None

        if not self._layer.currentText():
            show_info("No layer selected")
            return None

        kwargs = {
            "lp_filter_resolution_ang": self.lp_filter_resolution,
            "hp_filter_resolution_ang": self.hp_filter_resolution,
            "pixel_size": self.pixel_size,
        }
        filtered_image, mask = self.handle_filter(
            self.layer,
            self.filter_2d,
            **kwargs,
        )
        if filtered_image is None:
            return None

        # self.layer.visible = False
        if self.show_mask:
            self.napari_viewer.add_image(
                mask,
                name=f"MASK LP {int(self.lp_filter_resolution)} HP {int(self.hp_filter_resolution)} AP {self.pixel_size} - {self.layer.name}",
            )


        image = napari.layers.Image(
            filtered_image,
            name=f"LP {int(self.lp_filter_resolution)} HP {int(self.hp_filter_resolution)} AP {self.pixel_size} LIVE {self.filter_2d}  - {self.layer.name}",
            metadata=self.layer.metadata.copy(),
        )


        FilterImageWidget.update_links(self.layer, general.get_layer_id(self.napari_viewer, image))

        image.contrast_limits_range = [
            filtered_image.min(),
            filtered_image.max(),
        ]
        image.contrast_limits = [-3, 3]

        self.napari_viewer.layers.insert(
            self.napari_viewer.layers.index(self.layer) + 1, image
        )

        if self.filter_2d:
            connect2 = self.napari_viewer.dims.events.current_step.connect(
                lambda *x, new_layer=image, old_layer=self.layer, filter_kwargs=kwargs: self._filter_layer(
                    new_layer=new_layer, old_layer=old_layer, **filter_kwargs
                )
            )
            connect1 = image.events.visible.connect(
                lambda *x, new_layer=image, old_layer=self.layer, filter_kwargs=kwargs: self._filter_layer(
                    new_layer=new_layer, old_layer=old_layer, **filter_kwargs
                )
            )
            image.connect1 = connect1
            image.connect2 = connect2

    def _filter_layer(self, new_layer, old_layer, **filter_kwargs):
        if (
            new_layer not in self.napari_viewer.layers
            or old_layer not in self.napari_viewer.layers
        ):
            new_layer.events.disconnect(new_layer.connect1)
            self.napari_viewer.dims.events.current_step.disconnect(
                new_layer.connect2
            )
            return

        if new_layer.visible:
            filtered_image, _ = self.handle_filter(
                old_layer,
                True,
                **filter_kwargs,
            )
            new_layer.data[...] = filtered_image
            new_layer.contrast_limits_range = [
                filtered_image.min(),
                filtered_image.max(),
            ]
            new_layer.contrast_limits = [
                -3,
                3,
            ]
            with new_layer.events.visible.blocker():
                new_layer.visible = True

    def handle_filter(self, layer, filter_2d, **kwargs):
        if filter_2d:
            slice_axis = self.napari_viewer.dims.order[0]
            slice_index = self.napari_viewer.dims.current_step[slice_axis]
            slc = [slice(None)] * len(layer.data.shape)
            slc[slice_axis] = slice_index
            data = layer.data[tuple(slc)]
        else:
            data = layer.data
        try:
            filtered_image, mask = filters.bandpass_filter(
                data,
                log=show_info,
                **kwargs,
            )
            filtered_image = (
                filtered_image - np.mean(filtered_image)
            ) / np.std(filtered_image)
        except TypeError:
            return None, None

        return filtered_image, mask

    @Slot(str)
    def _update_pixel_size(self, layer_name):
        try:
            layer = self.napari_viewer.layers[layer_name]
        except KeyError:
            return

        try:
            pixel_spacing = layer.metadata["pixel_spacing"]
        except KeyError:
            self.pixel_size = -1
        else:
            self.pixel_size = pixel_spacing

        try:
            is_2d_stacked = layer.metadata["is_2d_stack"]
        except KeyError:
            self.filter_2d = False
            self._filter_2d.setEnabled(False)
        else:
            self.filter_2d = is_2d_stacked
            self._filter_2d.setEnabled(is_2d_stacked)

    @Slot(object)
    def _update_combo(self, *_):
        layer_names = sorted(
            entry.name
            for entry in self.napari_viewer.layers
            if isinstance(entry, napari.layers.Image)
        )
        current_text = self._layer.currentText()

        self._layer.currentTextChanged.disconnect(self._update_pixel_size)
        self._layer.clear()
        self._layer.addItems(layer_names)
        self._layer.setCurrentText(current_text)
        self._layer.currentTextChanged.connect(self._update_pixel_size)
        self._layer.currentTextChanged.emit(self._layer.currentText())

    @property
    def layer(self):
        return self.napari_viewer.layers[self._layer.currentText()]

    @property
    def show_mask(self):
        return self._show_mask.isChecked()

    @property
    def lp_filter_resolution(self):
        return float(self._lp_filter_resolution.text())

    @property
    def hp_filter_resolution(self):
        return float(self._hp_filter_resolution.text())

    @property
    def pixel_size(self):
        return float(self._pixel_size.text())

    @pixel_size.setter
    def pixel_size(self, value):
        self._pixel_size.setText(str(value))

    @property
    def filter_2d(self):
        return self._filter_2d.isChecked()

    @filter_2d.setter
    def filter_2d(self, value):
        return self._filter_2d.setChecked(value)



# """
# This module is an example of a barebones QWidget plugin for napari
#
# It implements the Widget specification.
# see: https://napari.org/plugins/guides.html?#widgets
#
# Replace code below according to your needs.
# """
from typing import TYPE_CHECKING

# from qtpy.QtWidgets import QHBoxLayout, QPushButton, QWidget

if TYPE_CHECKING:
    pass


# def example_magic_widget(img_layer: "napari.layers.Points"):
#    print(f"you have selected {img_layer}")
#
#
#
#
# class ExampleQWidget(QWidget):
#    # your QWidget.__init__ can optionally request the napari viewer instance
#    # in one of two ways:
#    # 1. use a parameter called `napari_viewer`, as done here
#    # 2. use a type annotation of 'napari.viewer.Viewer' for any parameter
#    def __init__(self, napari_viewer):
#        super().__init__()
#        self.viewer = napari_viewer
#
#        btn = QPushButton("Click me!")
#        btn.clicked.connect(self._on_click)
#
#        self.setLayout(QHBoxLayout())
#        self.layout().addWidget(btn)
#
#    def _on_click(self):
#        print("napari has", len(self.viewer.layers), "layers")
#
#
# @magic_factory
# def example_magic_widget(img_layer: "napari.layers.Image"):
#    print(f"you have selected {img_layer}")
#
#
## Uses the `autogenerate: true` flag in the plugin manifest
## to indicate it should be wrapped as a magicgui to autogenerate
## a widget.
# def example_function_widget(img_layer: "napari.layers.Image"):
#    print(f"you have selected {img_layer}")



import glob
import os
import typing

import matplotlib.cm as mcm
import numpy as np
import numpy.typing as npt
import pandas as pd

from .._utils import general
from . import io_utils as coordsio
from .interface import NapariLayerData
from .io_utils import MAX_LAYER_NAME

if typing.TYPE_CHECKING:
    pass


def write(path: "os.PathLike", output_df: pd.DataFrame):
    output_df.to_pickle(path)


def read(path: "os.PathLike") -> pd.DataFrame:
    """
    Read a tloc conform file into memory.
    tlpkl files need to have their coordinates for the Z and X dimension inverted.

    :param path: Path of the file to read information from.
    :type path: Path object or str

    :return: DataFrame containing the coordinates, box size, metrics, and cluster size
    :rtype: Pandas DataFrame
    """
    pandas_data: pd.DataFrame = pd.read_pickle(path)

    pandas_data["X"] = pandas_data["X"]
    pandas_data["Y"] = pandas_data["Y"]
    pandas_data["Z"] = pandas_data["Z"]

    return pandas_data


def get_valid_extensions():
    return ["tloc"]


def _get_meta_idx():
    return ["metric", "size"]


def _get_feature_idx():
    return ["grp_idx"]

def has_shapes(path: os.PathLike) -> bool:
    return False

def to_napari(
    path: os.PathLike | list[os.PathLike],
) -> "list[NapariLayerData]":
    file_name: str

    if not isinstance(path, list):
        path = sorted(glob.glob(path))  # type: ignore

    output_dfs = []
    for file_name in path:  # type: ignore
        input_df = read(file_name)
        r = coordsio.to_napari_coordinates(
            path=path,
            read_func=read,
            prepare_napari_func=_prepare_napari_box,
            meta_columns=_get_meta_idx(),
            feature_columns=_get_feature_idx(),
            valid_extensions=get_valid_extensions(),
        )

        data, kwargs, napari_type = r[0]
        metadata = kwargs["metadata"]
        features = pd.DataFrame(kwargs["features"])
        for cluster_id, cluster_df in features.groupby("grp_idx", sort=False):
            new_kwargs = kwargs.copy()
            new_metadata = {}
            for key, value in metadata.items():
                if isinstance(key, int):
                    if key in set(cluster_df.index):
                        new_metadata[key] = value
                else:
                    new_metadata[key] = value

            path = input_df.attrs["references"][cluster_id]
            if len(file_name) >= MAX_LAYER_NAME + 3:
                name = f"...{path[-MAX_LAYER_NAME:]}"  # type: ignore
            else:
                name = path  # type: ignore

            new_metadata["is_2d_stack"] = False
            new_metadata["references"] = input_df.attrs
            new_metadata["predicted_class"] = cluster_id

            new_kwargs["name"] = name
            new_kwargs["metadata"] = new_metadata
            new_kwargs["features"] = cluster_df.drop(
                columns=["grp_idx"]
            ).reset_index(drop=True)
            output_dfs.append(
                (
                    data.iloc[cluster_df.index].reset_index(drop=True),
                    new_kwargs,
                    napari_type,
                )
            )

    colors = mcm.get_cmap("gist_rainbow")
    n_layers = np.maximum(len(output_dfs), 2)  # Avoid zero division

    for cidx, (data, kwargs, _) in enumerate(output_dfs):
        metadata = kwargs["metadata"]
        for idx in range(int(data[["x", "y", "z"]].max().max().round(0)) + 1):
            idx_view_df = kwargs["features"].loc[data["x"].round(0) == idx, :]
            metadata[idx] = {
                "path": kwargs["metadata"]["original_path"],
                "name": "slice",
                "write": None,
            }
            metadata[idx].update(
                {
                    f"{entry}_{'min' if 'min' in func.__name__ else 'max'}": func(
                        idx_view_df[entry]
                    )
                    for func in [general.get_min_floor, general.get_max_ceil]
                    for entry in _get_meta_idx()
                }
            )

        cur_color = colors(cidx / (n_layers - 1))
        kwargs["edge_color"] = [cur_color]

    return output_dfs


# def to_napari_old(
#    path: os.PathLike | list[os.PathLike],
# ) -> "list[tuple[npt.ArrayLike, dict[str, typing.Any], str]]":
#    print("OLD")
#    input_df: pd.DataFrame
#    name: str
#    features: dict[str, typing.Any]
#
#    if not isinstance(path, list):
#        path = sorted(glob.glob(path))  # type: ignore
#
#    output_dfs = []
#    for file_name in path:  # type: ignore
#        input_df = read(file_name)
#        napari_df = _prepare_napari(input_df)
#        for cluster_id, cluster_df in napari_df.groupby("grp_idx", sort=False):
#            path = input_df.attrs["references"][cluster_id]
#            if len(path) >= MAX_LAYER_NAME + 3:
#                name = f"...{path[-MAX_LAYER_NAME:]}"  # type: ignore
#            else:
#                name = path  # type: ignore
#            output_dfs.append(
#                (
#                    cluster_id,
#                    file_name,
#                    name,
#                    cluster_df,
#                    input_df.attrs,
#                )
#            )
#
#    colors = mcm.get_cmap("gist_rainbow")
#    n_layers = np.maximum(len(output_dfs), 2)  # Avoid zero division
#
#    output_layers = []
#    for idx, (
#        cluster_id,
#        file_name,
#        cluster_name,
#        cluster_df,
#        attrs,
#    ) in enumerate(output_dfs):
#        cur_color = colors(idx / (n_layers - 1))
#        metadata = {
#            "input_attrs": attrs,
#            "predicted_class": cluster_id,
#            "set_lock": True,
#        }
#        for idx in range(
#            int(cluster_df[["x", "y", "z"]].max().max().round(0)) + 1
#        ):
#            idx_view_df = cluster_df.loc[cluster_df["x"].round(0) == idx, :]
#            metadata[idx] = {
#                "path": file_name,
#                "name": "slice",
#                "write": None,
#            }
#            metadata[idx].update(
#                {
#                    f"{entry}_{'min' if 'min' in func.__name__ else 'max'}": func(
#                        idx_view_df[entry]
#                    )
#                    for func in [general.get_min_floor, general.get_max_floor]
#                    for entry in _get_meta_idx()
#                }
#            )
#        features = {
#            entry: cluster_df[entry].to_numpy()
#            for entry in _get_meta_idx() + _get_hidden_meta_idx()
#        }
#        kwargs = {
#            "edge_color": [cur_color],
#            "face_color": "transparent",
#            "symbol": "disc",
#            "edge_width": 2,
#            "edge_width_is_relative": False,
#            "size": cluster_df["boxsize"],
#            "out_of_slice_display": True,
#            "opacity": 0.5,
#            "name": cluster_name,
#            "metadata": metadata,
#            "features": features,
#        }
#        output_layers.append(
#            (cluster_df[_get_3d_coords_idx()], kwargs, "points")
#        )
#
#    return output_layers
#
#


def _format_tloc(
    coordinates: pd.DataFrame,
    box_size: npt.ArrayLike,
    features: pd.DataFrame,
    metadata: dict[str, typing.Any],
    **kwargs,
) -> pd.DataFrame:
    column_names = [
        "X",
        "Y",
        "Z",
        "predicted_class",
        "size",
        "metric_best",
        "width",
        "height",
        "depth",
    ]

    data_df = pd.DataFrame(columns=column_names)
    data_df["X"] = coordinates[..., 2]
    data_df["Y"] = coordinates[..., 1]
    data_df["Z"] = coordinates[..., 0]

    data_df["size"] = features["size"].to_numpy()
    data_df["metric_best"] = features["metric"].to_numpy()

    data_df["predicted_class"] = metadata["predicted_class"]
    data_df["width"] = box_size
    data_df["height"] = box_size
    data_df["depth"] = box_size
    data_df.attrs = metadata["references"]

    return data_df


def _write_tloc(path: os.PathLike, df: pd.DataFrame, **kwargs):
    write(path, df)


def from_napari(
    path: os.PathLike,
    layer_data: list[NapariLayerData],
    suffix: str,
    filament_spacing: int,
):
    path = coordsio.from_napari(
        path=path,
        layer_data=layer_data,
        write_func=_write_tloc,
        format_func=_format_tloc,
        suffix=suffix,
        filament_spacing=filament_spacing,
    )
    return path


# def from_napari_old(
#    path: os.PathLike,
#    layer_data: typing.Any,
# ):
#    column_names = [
#        "X",
#        "Y",
#        "Z",
#        "predicted_class",
#        "size",
#        "metric_best",
#        "width",
#        "height",
#        "depth",
#    ]
#    output_dfs = []
#    for coords, meta, _ in layer_data:
#        shown_mask = meta["shown"]
#        data_df = pd.DataFrame(columns=column_names)
#        features = meta["features"]
#        data_df["X"] = coords[shown_mask, 2]
#        data_df["Y"] = coords[shown_mask, 1]
#        data_df["Z"] = coords[shown_mask, 0]
#        data_df["predicted_class"] = meta["metadata"]["predicted_class"]
#        data_df["size"] = features.loc[shown_mask, "size"].to_numpy()
#        data_df["metric_best"] = features.loc[shown_mask, "metric"].to_numpy()
#        data_df["width"] = meta["size"][shown_mask, 2]
#        data_df["height"] = meta["size"][shown_mask, 1]
#        data_df["depth"] = meta["size"][shown_mask, 0]
#        data_df.attrs = meta["metadata"]["input_attrs"]
#        output_dfs.append(data_df)
#
#    output_df = pd.concat(output_dfs, ignore_index=True, axis=0)
#    write(path, output_df)
#
#    return path


def _prepare_napari_box(
    input_df: pd.DataFrame,
) -> pd.DataFrame:
    output_data: pd.DataFrame = pd.DataFrame(
        columns=_get_3d_coords_idx()
        + _get_meta_idx()
        + _get_hidden_meta_idx()
        + _get_util_idx()
    )

    output_data["x"] = input_df["Z"]
    output_data["y"] = input_df["Y"]
    output_data["z"] = input_df["X"]
    output_data["metric"] = input_df["metric_best"]
    output_data["size"] = input_df["size"]
    output_data["boxsize"] = input_df[["height", "width", "depth"]].mean(
        axis=1
    )
    output_data["grp_idx"] = input_df["predicted_class"]

    return output_data


def _get_3d_coords_idx():
    return ["x", "y", "z"]


def _get_hidden_meta_idx():
    return []


def _get_util_idx():
    return ["grp_idx"]



import glob
import io
import itertools
import os
import pathlib
import typing
import warnings
from collections.abc import Callable
from typing import Protocol, Union, Dict

import matplotlib.pyplot as plt
import napari
import numpy as np
import numpy.typing as npt
import pandas as pd

from .._qt import OrganizeBox as orgbox

from .interface import NapariLayerData, NapariMetaData


class IllegalFormatException(Exception):
    pass


with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        r"The numpy.array_api submodule is still experimental. See NEP 47.",
    )
    from numpy.array_api._array_object import Array

MAX_LAYER_NAME: int = 30
PROXY_THRESHOLD_GB: float = (
    float(os.environ["BOXMANAGER_PROXY_GB"])
    if "BOXMANAGER_PROXY_GB" in os.environ
    else 2
)


class FormatFunc(Protocol):
    def __call__(
        self,
        coordinates: pd.DataFrame,
        boxsize: npt.ArrayLike,
        features: pd.DataFrame,
        metadata: dict,
        filament_spacing: int,
    ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        ...


def _split_filaments(data: pd.DataFrame) -> typing.List[pd.DataFrame]:
    vert = None
    if 'filament_vertices' in data.attrs:
        vert = data.attrs['filament_vertices']
    filament_ids = np.unique(data["fid"])
    filaments = []
    for id in filament_ids:
        mask = data["fid"] == id
        filament = data[mask]

        if vert:
            mask = vert["_filamentid"] == id
            vert_fil = vert[mask]
            filament.attrs['filament_vertices'] = vert_fil

        filaments.append(filament)
    return filaments


def _prepare_coords_df(
    path: list[os.PathLike],
    read_func: Callable[[os.PathLike], pd.DataFrame],
    prepare_napari_func: Callable,
    meta_columns: typing.List[str] = [],
) -> tuple[typing.List[pd.DataFrame], dict, bool]:

    data_df: list[pd.DataFrame] = []
    metadata: dict = {}
    is_3d = True
    is_filament = False
    for idx, entry in enumerate(path):
        input_data = read_func(entry)

        box_napari_data = pd.DataFrame(columns=meta_columns)

        if input_data is not None:
            box_napari_data = prepare_napari_func(input_data)

            if "x" not in box_napari_data:
                is_3d = False
                box_napari_data["x"] = idx
            if "fid" in box_napari_data:
                is_filament = True
                box_napari_data = _split_filaments(box_napari_data)
        checkbox = None
        if is_filament:
            if len(box_napari_data) == 0 or np.all(
                [b.empty for b in box_napari_data]
            ):
                checkbox = True
        elif box_napari_data.empty:
            checkbox = True

        if checkbox:
            # needs initialization if input file is empty
            min_max_data = set()
        elif is_filament:
            data_df.extend(box_napari_data)
            min_max_data = pd.concat(box_napari_data)
        else:
            min_max_data = box_napari_data
            data_df.append(box_napari_data)

        metadata[idx] = {}
        metadata[idx]["path"] = entry
        metadata[idx]["name"] = os.path.basename(entry)
        metadata[idx]["write"] = checkbox

        try:
            metadata[idx].update(
                {
                    f"{col}_{func.__name__}": func(min_max_data[col])
                    for func in [min, max]
                    for col in meta_columns
                    if col in min_max_data
                }
            )
        except ValueError:
            pass
    metadata["is_filament_layer"] = is_filament

    return data_df, metadata, is_3d


def is_filament_layer(
    layer_data: typing.Union[NapariLayerData, list[NapariLayerData]]
) -> bool:
    def check(ldat: NapariLayerData) -> bool:
        is_filament = (
            "is_filament_layer" in ldat[1]["metadata"]
            and ldat[1]["metadata"]["is_filament_layer"]
        )
        return is_filament

    if isinstance(layer_data, list):
        if np.sum([check(layer) for layer in layer_data]) not in (
            0,
            len(layer_data),
        ):
            raise IllegalFormatException(
                "Only layers of the same type (either particle or filament can be saved."
            )
        is_filament = check(layer_data[0])
    else:
        is_filament = check(layer_data)
    return is_filament


def get_layers_name(path: os.PathLike | list[os.PathLike]):

    if not isinstance(path, list):
        if len(path) >= MAX_LAYER_NAME + 3:
            name = f"...{path[-MAX_LAYER_NAME:]}"  # type: ignore
        else:
            name = path  # type: ignore
    else:
        name = (os.path.splitext(path[0])[1]).upper()[1:]

    return name


def _to_napari_filament(input_df: list[pd.DataFrame], coord_columns, is_3d):
    # boxsize_ = [np.mean(fil['boxsize']) for fil in input_df]

    color = []
    boxsize = []
    total = 0

    for fil in input_df:
        bs = np.mean(fil["boxsize"])
        c = np.random.choice(range(256), size=3)
        total += len(fil)
        color.extend([c] * len(fil))
        boxsize.extend([bs] * len(fil))
    input_df = pd.concat([fil[coord_columns] for fil in input_df])
    color = [(r, g, b, 255) for r, g, b in color]

    kwargs: NapariMetaData = {
        "edge_color": color,
        "face_color": "transparent",
        "symbol": "disc",
        "edge_width": 0.05,
        "edge_width_is_relative": True,
        "size": np.average(boxsize),
        "out_of_slice_display": True if is_3d else False,
        "opacity": 0.8,
    }
    dat = input_df
    layer_type = "points"

    return dat, kwargs, layer_type

def _to_napari_filament_shapes(input_df: list[pd.DataFrame], coord_columns, is_3d):
    boxsize = np.median([np.mean(fil["boxsize"]) for fil in input_df])
    input_df = np.array([fil[coord_columns].to_numpy() for fil in input_df])

    kwargs = {
        "edge_color": ["red"],
        "face_color": "transparent",
        "edge_width": boxsize,
        "opacity": 0.4,
        "shape_type": "path",
    }
    dat = input_df
    layer_type = "shapes"

    return dat, kwargs, layer_type


def _to_napari_particle(input_df, coord_columns, is_3d):
    input_df = pd.concat(input_df, ignore_index=True)

    kwargs: NapariMetaData = {
        "edge_color": "red",
        "face_color": "transparent",
        "symbol": "disc",
        "edge_width": 0.05,
        "edge_width_is_relative": True,
        "size": np.average(input_df["boxsize"]),
        "out_of_slice_display": True if is_3d else False,
        "opacity": 0.8,
    }
    dat = input_df[coord_columns]
    layer_type = "points"

    return dat, kwargs, layer_type


def to_napari_image(
    path: os.PathLike | list[os.PathLike],
    load_image: typing.Callable[[str], np.array],
    get_pixel_size: typing.Callable[[str], float],
    do_normalize: bool = True,
) -> "list[tuple[npt.ArrayLike, dict[str, typing.Any], str]]":

    is_2d_stack = isinstance(path, list) or "*" in path
    name = get_layers_name(path)

    if not isinstance(path, list):
        original_path = path
        path = sorted(glob.glob(path))  # type: ignore
    else:
        original_path = (
            f"{os.path.dirname(path[0])}/*{os.path.splitext(path[0])[1]}"
        )

    # arrays = []
    voxel_size = 1
    metadata: dict = {
        "pixel_spacing": voxel_size,
        "original_path": original_path,
    }
    metadata["pixel_spacing"] = get_pixel_size(path[0])

    file_size = (
        sum(os.stat(file_name).st_size for file_name in path) / 1024**3
    )
    if len(path) > 1 and file_size > PROXY_THRESHOLD_GB:
        data = LoaderProxy(path, load_image, do_normalize)
    else:
        data_list = []
        for file_name in path:
            tmp_data = load_image(file_name)
            if do_normalize:
                tmp_data = (tmp_data - np.mean(tmp_data)) / np.std(tmp_data)
            data_list.append(tmp_data)
        data = np.squeeze(np.stack(data_list))

    metadata["is_3d"] = len(path) == 1 and data.ndim == 3
    metadata["is_2d_stack"] = is_2d_stack

    if (metadata["is_3d"] or metadata["is_2d_stack"]) and data.ndim == 2:
        data = np.expand_dims(data, axis=0)

    if not metadata["is_3d"]:
        for idx, file_name in enumerate(path):
            metadata[idx] = {}
            metadata[idx]["path"] = file_name
            metadata[idx]["name"] = os.path.basename(file_name)

    add_kwargs = {"metadata": metadata, "name": name}

    layer_type = "image"  # optional, default is "image"
    return [(data, add_kwargs, layer_type)]



def to_napari_coordinates(
    path: os.PathLike | list[os.PathLike],
    read_func: Callable[[os.PathLike], pd.DataFrame],
    prepare_napari_func: Callable,
    meta_columns: typing.List[str] = [],
    feature_columns: typing.List[str] = [],
    valid_extensions: typing.List[str] = [],
    make_filament_shape_layer: bool = False,
) -> "list[NapariLayerData]":

    input_df_list: list[pd.DataFrame]
    features: dict[str, typing.Any]

    is_2d_stack = isinstance(path, list) or "*" in path
    orgbox_meta = orgbox.get_metadata(path)


    layer_name = get_layers_name(path)


    if not isinstance(path, list):
        path = sorted(glob.glob(path))  # type: ignore

    input_df_list, metadata, is_3d = _prepare_coords_df(
        path,
        read_func=read_func,
        prepare_napari_func=prepare_napari_func,
        meta_columns=meta_columns,
    )
    metadata.update(orgbox_meta)
    metadata["is_2d_stack"] = is_2d_stack
    metadata["ignore_idx"] = feature_columns

    features = {}
    for entry in feature_columns + meta_columns:
        if metadata["is_filament_layer"]:
            for fil in input_df_list:
                if entry not in fil:
                    continue
                entry_data = fil[entry].to_numpy()
                if make_filament_shape_layer:
                    # instead per filament box, in case of shapes layer it is per shape
                    entry_data = np.unique(entry_data)
                if entry in features:
                    features[entry] = np.concatenate(
                        [features[entry], entry_data]
                    )
                else:
                    features[entry] = entry_data
        else:
            all = pd.concat(input_df_list)
            if entry in all:
                features[entry] = all[entry].to_numpy()

    if is_2d_stack or is_3d:
        # Happens for --stack option and '*.ext'
        coord_columns = ["x", "y", "z"]
    else:
        coord_columns = ["y", "z"]

    if metadata["is_filament_layer"]:
        if make_filament_shape_layer:
            dat, kwargs, layer_type = _to_napari_filament_shapes(
                input_df_list, coord_columns, is_3d
            )
        else:
            dat, kwargs, layer_type = _to_napari_filament(
                input_df_list, coord_columns, is_3d
            )
    else:
        dat, kwargs, layer_type = _to_napari_particle(
            input_df_list, coord_columns, is_3d
        )
    kwargs["name"] = layer_name
    kwargs["metadata"] = metadata
    kwargs["features"] = features

    return [(dat, kwargs, layer_type)]


def _generate_output_filename(
    orignal_filename: str, output_path: os.PathLike, suffix=""
):
    if not os.path.isdir(output_path):
        dirname = os.path.dirname(output_path)
        file_base, extension = os.path.splitext(os.path.basename(output_path))
        if not extension:  # in case '.box' is provided as output path.
            file_base, extension = extension, file_base
    else:
        extension = suffix
        dirname = output_path
        file_base = os.path.splitext(os.path.basename(orignal_filename))[0]

    output_file = pathlib.Path(dirname, file_base + extension)
    return output_file


def resample_filament(
    filament: pd.DataFrame,
    new_distance: float,
    coordinate_columns,
    constant_columns=[],
    other_interpolation_col=[],
):
    if len(filament) <= 1:
        return filament
    import numpy as np
    from scipy.interpolate import interp1d

    new_distance = np.sqrt(new_distance**2 + new_distance**2)

    constants = {}
    for col in constant_columns:
        if col in filament:
            constants[col] = filament[col][0]

    sqsum = 0
    for col in coordinate_columns:
        coord_dat = filament[col].to_list()
        sqsum = sqsum + np.ediff1d(coord_dat, to_begin=0) ** 2

    distance_elem = np.cumsum(np.sqrt(sqsum))
    total_elength = distance_elem[-1]

    distance = distance_elem / distance_elem[-1]  # norm to 1
    interpolators = []
    for col in coordinate_columns + other_interpolation_col:
        interp = interp1d(distance, filament[col].to_list())
        interpolators.append(interp)

    num = int(total_elength / (new_distance)) + 1

    alpha = np.linspace(0, 1, num)

    interpolated = []
    for interp in interpolators:
        interpolated.append(interp(alpha))

    ####
    # Generate new boxes
    ####
    new_boxes = {}
    for col_i, col in enumerate(coordinate_columns + other_interpolation_col):
        new_boxes[col] = interpolated[col_i]

    for col in constant_columns:
        new_boxes[col] = [constants[col]] * len(interpolated[0])
    new_fil = pd.DataFrame(new_boxes)

    return new_fil


def _write_particle_data(
    path: os.PathLike,
    data: NapariLayerData,
    meta: NapariLayerData,
    format_func: FormatFunc,
    write_func: Callable[[os.PathLike, pd.DataFrame, ...], typing.Any],
    suffix: str = "",
    filament_spacing: float = 0,
):
    if data.shape[1] == 2:
        data = np.insert(data, 0, 0, axis=1)

    kwargs = {}

    if "shown" in meta:
        mask = meta["shown"]
    else:
        # For filaments
        mask = np.ones(len(data), dtype=int) == 1

    coordinates = data[mask]

    if "size" in meta:
        boxsize = meta["size"][mask]
    else:
        # For filaments
        boxsize = np.array(meta["edge_width"])[mask]
    export_data = {}
    try:
        is_2d_stacked = meta["metadata"]["is_2d_stack"]
    except KeyError:
        is_2d_stacked = False

    if is_2d_stacked:
        for z in meta["metadata"]:
            if not isinstance(z, int) or meta["metadata"][z]["write"] is False:
                continue
            mask = coordinates[:, 0] == z
            if np.sum(mask) == 0 and meta["metadata"][z]["write"] is not True:
                continue

            filename = meta["metadata"][z]["name"]
            kwargs["image_name"] = filename
            output_file = _generate_output_filename(
                orignal_filename=filename, output_path=path, suffix=suffix
            )

            d = format_func(
                coordinates=coordinates[mask, 1:],
                box_size=boxsize[mask],
                features=meta["features"].loc[mask, :],
                metadata=meta["metadata"],
                filament_spacing=filament_spacing,
            )
            export_data[output_file] = (d, {})

    else:
        filename = meta["metadata"]["original_path"]
        output_file = _generate_output_filename(
            orignal_filename=filename, output_path=path, suffix=suffix
        )
        empty_slices = []
        slices_with_coords = np.unique(coordinates[:, 0]).tolist()
        for z in meta["metadata"]:
            if (
                not isinstance(z, int)
                or meta["metadata"][z]["write"] is not True
            ):
                continue
            if z in slices_with_coords:
                continue
            empty_slices.append(z)

        export_data[output_file] = (
            format_func(
                coordinates=coordinates,
                box_size=boxsize,
                features=meta["features"].loc[mask, :],
                metadata=meta["metadata"],
                filament_spacing=filament_spacing,
            ),
            {"empty_slices": empty_slices},
        )

    for outpth in export_data:
        df = export_data[outpth][0]
        write_func(outpth, df, **export_data[outpth][1])
    last_file = outpth
    return str(last_file)


def convert_shape_filament_layer_to_boxlayer(
    data: list[np.array], meta: dict
) -> (np.array, dict):
    repeat = []
    for fil in data:
        repeat.append(len(fil))
    meta["features"] = meta["features"].loc[
        meta["features"].index.repeat(repeat)
    ]

    meta["edge_width"] = list(
        itertools.chain.from_iterable(
            [
                [meta["edge_width"][filid]] * len(d)
                for filid, d in enumerate(data)
            ]
        )
    )

    meta["features"]["fid"] = list(
        itertools.chain.from_iterable(
            [[filid] * len(d) for filid, d in enumerate(data)]
        )
    )

    data = np.concatenate(data)

    return data, meta


def from_napari(
    path: os.PathLike,
    layer_data: list[NapariLayerData],
    format_func: FormatFunc,
    write_func: Callable[
        [os.PathLike, pd.DataFrame | list[pd.DataFrame]], typing.Any
    ],
    suffix: str = "",
    filament_spacing: int = 0,
) -> os.PathLike:

    last_file = ""
    for data, meta, layer in layer_data:
        is_filament_data = is_filament_layer(layer_data)
        if isinstance(data, list):
            data, meta = convert_shape_filament_layer_to_boxlayer(data, meta)

        if is_filament_data:

            if "edge_width" not in meta:
                boxsize = meta["features"]["boxsize"]
                meta["edge_width"] = boxsize

            fid = meta["features"]["fid"]

            fid = np.array(fid, dtype=int)

            data_list = np.append(data, np.atleast_2d(np.array(fid)).T, axis=1)
            last_file = _write_particle_data(
                path,
                data_list,
                meta,
                format_func,
                write_func,
                suffix,
                filament_spacing,
            )
        else:
            last_file = _write_particle_data(
                path,
                data,
                meta,
                format_func,
                write_func,
                suffix,
                filament_spacing,
            )

    return str(last_file)


class LoaderProxy(Array):
    def __init__(self, files, reader_func, do_normalize):
        self.reader_func = reader_func
        self.files = files
        self.do_normalize = do_normalize
        if len(self.files) == 0:
            raise AttributeError("Cannot provide empty files list")

        self._array = None
        self._shape = self.load_image(0).shape

    @property
    def shape(self):
        return (len(self.files), *self._shape)

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        return obj

    def load_image(self, index) -> Array:
        data = self.reader_func(self.files[index])
        self._array = np.empty((1, 1, 1), dtype=data.dtype)
        if self.do_normalize:
            return (data - np.mean(data)) / np.std(data)
        else:
            return data

    def __len__(self):
        return len(self.files)

    def __iter__(self):
        return (self[idx] for idx in range(len(self.files)))

    def __getitem__(self, key):
        try:
            super()._validate_index(key, None)
        except TypeError:
            super()._validate_index(key)

        if isinstance(key, Array):
            key._array

        try:
            _key = key[0]
        except TypeError:
            _key = key

        if isinstance(_key, (int, np.integer)):
            return self.load_image(_key)
        else:
            return self.get_dummy_image()

    def __copy__(self):
        return LoaderProxy(self.files, self.reader_func)

    def __deepcopy__(self, _):
        return self.__copy__()

    def get_dummy_image(self):
        size = self.shape[-1]
        fig = plt.figure(figsize=(size, size), dpi=1)
        new_shape = (int(fig.bbox.bounds[3]), int(fig.bbox.bounds[2]), -1)
        plt.text(
            0.5,
            0.5,
            "\U00002639",
            va="center_baseline",
            ha="center",
            fontsize=size * 50,
        )
        plt.axis("off")

        io_buf = io.BytesIO()
        fig.savefig(io_buf, format="raw")
        io_buf.seek(0)
        img_arr = np.reshape(
            np.frombuffer(io_buf.getvalue(), dtype=np.uint8),
            newshape=new_shape,
        )[..., 0]
        io_buf.close()
        plt.close(fig)
        return img_arr



# import numpy as np
#
# from box_manager import ExampleQWidget, example_magic_widget
#
#
# make_napari_viewer is a pytest fixture that returns a napari viewer object
# capsys is a pytest fixture that captures stdout and stderr output streams
# def test_example_q_widget(make_napari_viewer, capsys):
#    # make viewer and add an image layer using our fixture
#    viewer = make_napari_viewer()
#    viewer.add_image(np.random.random((100, 100)))
#
#    # create our widget, passing in the viewer
#    my_widget = ExampleQWidget(viewer)
#
#    # call our widget method
#    my_widget._on_click()
#
#    # read captured output and check that it's as we expected
#    captured = capsys.readouterr()
#    assert captured.out == "napari has 1 layers\n"
#
#
# def test_example_magic_widget(make_napari_viewer, capsys):
#    viewer = make_napari_viewer()
#    layer = viewer.add_image(np.random.random((100, 100)))
#
#    # this time, our widget will be a MagicFactory or FunctionGui instance
#    my_widget = example_magic_widget()
#
#    # if we "call" this object, it'll execute our function
#    my_widget(viewer.layers[0])
#
#    # read captured output and check that it's as we expected
#    captured = capsys.readouterr()
#    assert captured.out == f"you have selected {layer}\n"



import os

import mrcfile
import numpy as np

OUTPUT_DIR = os.path.dirname(__file__)
NDIM = 50
np.random.seed(42)

for mic_id in range(10):
    data = np.zeros((NDIM, NDIM), dtype=np.float32)
    coords = []
    for box_id in range(3):
        x_coord = int(np.random.random() * NDIM)
        y_coord = int(np.random.random() * NDIM)
        coords.append((y_coord, x_coord, 0, 0))
        data[x_coord, y_coord] = 1

    with mrcfile.new(
        os.path.join(OUTPUT_DIR, f"test_{mic_id}.mrc"), overwrite=True
    ) as new:
        new.set_data(data)

    for suffix in ("", "_centered", "_new"):
        if np.random.random() > 0.8:
            continue

        with open(f"test_{mic_id}{suffix}.box", "w") as write:
            write.write(
                "\n".join([" ".join(map(str, entry)) for entry in coords])
            )



import os

import numpy as np
import numpy.typing as npt
import pandas as pd

from . import io_utils as coordsio
from .interface import NapariLayerData


class BoxFileNumberOfColumnsError(pd.errors.IntCastingNaNError):
    pass


class UnknownFormatException(Exception):
    ...


DEFAULT_BOXSIZE: int = 10
DEFAULT_RESAMPLING_FILAMENT: int = -1  # 20% of boxsize


def get_valid_extensions():
    return ["box"]


def is_helicon_with_particle_coords(path):
    try:
        with open(path) as f:
            first_line = f.readline()
            f.close()
        return "#micrograph" in first_line
    except ValueError:
        return False


def read_boxfile(path: "os.PathLike") -> pd.DataFrame:
    names = ["x", "y", "box_x", "box_y"]
    box_data: pd.DataFrame = pd.read_csv(
        path,
        delim_whitespace=True,
        index_col=False,
        header=None,
        dtype=float,
        names=names,
        usecols=range(len(names)),
    )  # type: ignore
    try:
        box_data.astype(int)
    except pd.errors.IntCastingNaNError:
        raise BoxFileNumberOfColumnsError
    return box_data


def read_helicon_boxfile(path: "os.PathLike") -> pd.DataFrame:
    def get_first_and_last_coord(helix_line):
        if not helix_line.startswith("#helix"):
            raise ValueError("Line does not start with '#helix'")
        import re

        result = re.findall(r"\d+.\d*", helix_line)
        allnumbers = [float(r) for r in result]
        return allnumbers[:4]

    if os.stat(path).st_size != 0:
        split_indicis = []
        boxsize = 0
        index_first_helix = -1
        csvlines = None
        with open(path) as csvfile:
            csvlines = csvfile.readlines()
            helixlines_indicies = []
            for index, row in enumerate(csvlines):
                if row.startswith("#segment"):
                    boxsize = int(float(row.split()[2]))
                elif row.startswith("#helix"):
                    boxsize = int(float(row[(row.rfind(",") + 1) :]))
                    if index_first_helix == -1:
                        index_first_helix = index
                    else:

                        split_indicis.append(
                            index
                            - index_first_helix
                            - (len(split_indicis) + 1)
                        )
                    helixlines_indicies.append(index)

        coordinates = np.atleast_2d(np.genfromtxt(path))
        # coordinates_lowleftcorner = coordinates - boxsize / 2
        coord_filaments = np.split(coordinates, split_indicis)

        data_dict = {"x": [], "y": [], "box_x": [], "box_y": [], "fid": []}
        for filament_index, filament in enumerate(coord_filaments):
            first_and_last_coord = get_first_and_last_coord(
                csvlines[helixlines_indicies[filament_index]]
            )

            # first
            data_dict["x"].append(first_and_last_coord[0])
            data_dict["y"].append(first_and_last_coord[1])
            data_dict["box_x"].append(boxsize)
            data_dict["box_y"].append(boxsize)
            data_dict["fid"].append(filament_index + 1)

            # in between
            for coords in filament:
                if len(coords) > 0:
                    data_dict["x"].append(coords[0])
                    data_dict["y"].append(coords[1])
                    data_dict["box_x"].append(boxsize)
                    data_dict["box_y"].append(boxsize)
                    data_dict["fid"].append(filament_index + 1)

            # last
            data_dict["x"].append(first_and_last_coord[2])
            data_dict["y"].append(first_and_last_coord[3])
            data_dict["box_x"].append(boxsize)
            data_dict["box_y"].append(boxsize)
            data_dict["fid"].append(filament_index + 1)

        data_df = pd.DataFrame(data_dict)

        ## make lower left corner to be compatible with prepare function
        data_df["x"] = data_df["x"] - boxsize // 2
        data_df["y"] = data_df["y"] - boxsize // 2

        return data_df

    return None

def has_shapes(path: os.PathLike) -> bool:
    return False

def write_eman1_helicon(path: str, filaments: list[pd.DataFrame], **kwargs):

    import csv

    image_filename = "NA"
    if "image_name" in kwargs:
        image_filename = kwargs["image_name"]
    with open(path, "w", newline="", encoding="utf-8") as boxfile:
        boxwriter = csv.writer(
            boxfile,
            delimiter="\t",
            quotechar="|",
            quoting=csv.QUOTE_NONE,
            lineterminator="\n",
        )
        # micrograph: actin_cAla_1_corrfull.mrc
        # segment length: 384
        # segment width: 384

        if filaments is not None and len(filaments) > 0:

            boxsize = filaments[0]["boxsize"][0]

            boxwriter.writerow(["#micrograph: " + image_filename])
            boxwriter.writerow(["#segment length: " + str(int(boxsize))])
            boxwriter.writerow(["#segment width: " + str(int(boxsize))])

            for fil in filaments:
                if len(fil) > 0:
                    boxwriter.writerow(
                        [
                            "#helix: ("
                            + str(fil["x"][0] + boxsize / 2)
                            + ", "
                            + str(fil["y"][0] + boxsize / 2)
                            + "),"
                            + "("
                            + str(fil["x"][len(fil) - 1] + boxsize / 2)
                            + ", "
                            + str(fil["y"][len(fil) - 1] + boxsize / 2)
                            + "),"
                            + str(int(boxsize))
                        ]
                    )
                    for index, box in fil.iterrows():
                        boxwriter.writerow(
                            [
                                ""
                                + str(box["x"] + boxsize / 2)
                                + " "
                                + str(box["y"] + boxsize / 2)
                            ]
                        )


def read(path: "os.PathLike") -> pd.DataFrame:
    if is_helicon_with_particle_coords(path):
        box_data: pd.DataFrame = read_helicon_boxfile(path)
    else:
        box_data: pd.DataFrame = read_boxfile(path)
    return box_data


def to_napari(
    path: os.PathLike | list[os.PathLike],
) -> "list[NapariLayerData]":
    r = coordsio.to_napari_coordinates(
        path=path,
        read_func=read,
        prepare_napari_func=_prepare_napari_box,
        meta_columns=_get_meta_idx(),
        feature_columns=_get_feature_idx(),
        valid_extensions=get_valid_extensions(),
    )
    return r


def _get_meta_idx():
    return []


def _get_feature_idx():
    return ["fid", "boxsize"]


def _prepare_napari_box(
    input_df: pd.DataFrame,
) -> pd.DataFrame:
    """

    Parameters
    ----------
    input_df Raw data from read function

    Returns
    -------
    Dataframe with centered coordinates with box size. Returns a list of dataframes in case of filaments

    """
    output_data: pd.DataFrame = pd.DataFrame(
        columns=["y", "z"] + _get_meta_idx()
    )

    output_data["z"] = input_df["x"] + input_df["box_x"] // 2
    output_data["y"] = input_df["y"] + input_df["box_y"] // 2

    output_data["boxsize"] = np.maximum(
        input_df[["box_x", "box_y"]].mean(axis=1), DEFAULT_BOXSIZE
    ).astype(int)

    if "fid" in input_df:
        output_data["fid"] = input_df["fid"]

    return output_data


def _write_box(path: os.PathLike, df: pd.DataFrame, **kwargs):
    df[["x", "y", "boxsize", "boxsize"]].to_csv(
        path, sep=" ", index=None, header=None
    )


def _make_df_data_particle(
    coordinates: pd.DataFrame, box_size: npt.ArrayLike, **kwargs
) -> pd.DataFrame:
    data = {"x": [], "y": [], "boxsize": []}
    for (y, x), boxsize in zip(
        coordinates,
        box_size,
    ):
        data["x"].append(x - boxsize / 2)
        data["y"].append(y - boxsize / 2)
        data["boxsize"].append(boxsize)
    return pd.DataFrame(data)


def _make_df_data_filament(
    coordinates: pd.DataFrame,
    box_size: npt.ArrayLike,
    filament_spacing: int,
    **kwargs,
) -> list[pd.DataFrame]:
    data = {"x": [], "y": [], "fid": [], "boxsize": []}
    filaments = []

    for (y, x, fid), boxsize in zip(
        coordinates,
        box_size,
    ):
        if len(data["fid"]) > 0 and data["fid"][-1] != fid:
            filaments.append(pd.DataFrame(data))
            data = {"x": [], "y": [], "fid": [], "boxsize": []}

        data["x"].append(x - boxsize / 2)
        data["y"].append(y - boxsize / 2)
        data["fid"].append(fid)
        data["boxsize"].append(boxsize)
    filaments.append(pd.DataFrame(data))

    ## Resampling
    for index_fil, fil in enumerate(filaments):
        if not filament_spacing:
            distance = int(fil["boxsize"][0] * 0.2)
        else:
            distance = filament_spacing
        filaments[index_fil] = coordsio.resample_filament(
            fil,
            distance,
            coordinate_columns=["x", "y"],
            constant_columns=["boxsize"],
        )

    return filaments


def from_napari(
    path: os.PathLike,
    layer_data: list[NapariLayerData],
    suffix: str,
    filament_spacing: int,
):
    is_filament = coordsio.is_filament_layer(layer_data)
    if is_filament:
        format_func = _make_df_data_filament
        write_func = write_eman1_helicon
    else:
        format_func = _make_df_data_particle
        write_func = _write_box
    path = coordsio.from_napari(
        path=path,
        layer_data=layer_data,
        write_func=write_func,
        format_func=format_func,
        suffix=suffix,
        filament_spacing=filament_spacing,
    )

    return path



import os
import typing
import tifffile
import numpy as np
import pandas as pd
from .io_utils import to_napari_image

if typing.TYPE_CHECKING:
    import numpy.typing as npt

def has_shapes(path: os.PathLike) -> bool:
    return False

def load_image(path: str) -> np.ndarray:

    with tifffile.TiffFile(path) as tif:
        img = tif.pages[0].asarray()
        img = img.astype(np.float32)
        if len(img.shape) == 3:
            img = np.flip(img, 1)
        else:
            img = np.flip(img, 0)
        return img
    return None


def get_pixel_size(path: str) -> float:
    with tifffile.TiffFile(path) as tif:
        try:
            return tif.pages[0].tags['TVIPS'].value['PixelSizeX']*10
        except KeyError:
            print("Can't find pixelsize. Use default (1).")

    return 1


def to_napari(
    path: os.PathLike | list[os.PathLike],
) -> "list[tuple[npt.ArrayLike, dict[str, typing.Any], str]]":

    return to_napari_image(path, load_image=load_image, get_pixel_size=get_pixel_size)


def get_valid_extensions() -> typing.List[str]:
    return ["tif", "tiff"]


def from_napari(
    path: os.PathLike | list[os.PathLike] | pd.DataFrame,
    data: typing.Any,
    meta: dict,
):
    raise NotImplementedError



# """
# This module is an example of a barebones sample data provider for napari.
#
# It implements the "sample data" specification.
# see: https://napari.org/plugins/guides.html?#sample-data
#
# Replace code below according to your needs.
# """
# from __future__ import annotations
#
# import numpy
#
#
# def make_sample_data():
#    """Generates an image"""
#    return [(numpy.random.rand(512, 512), {}), (numpy.random.rand(20, 20), {})]



import os

import numpy.typing as npt
import pandas as pd

from . import io_utils as coordsio
from .interface import NapariLayerData


class BoxFileNumberOfColumnsError(pd.errors.IntCastingNaNError):
    pass


class UnknownFormatException(Exception):
    ...


DEFAULT_BOXSIZE: int = 10


def get_valid_extensions():

    return ["coords"]

def has_shapes(path: os.PathLike) -> bool:
    return False

def read(path: os.PathLike) -> pd.DataFrame:

    names = ["x", "y", "z"]
    box_data: pd.DataFrame = pd.read_csv(
        path,
        delim_whitespace=True,
        index_col=False,
        header=None,
        dtype=float,
        names=names,
        usecols=range(len(names)),
    )  # type: ignore
    try:
        box_data.astype(int)
    except pd.errors.IntCastingNaNError:
        raise BoxFileNumberOfColumnsError

    return box_data


def to_napari(
    path: os.PathLike | list[os.PathLike],
) -> "list[NapariLayerData]":

    return coordsio.to_napari_coordinates(
        path=path,
        read_func=read,
        prepare_napari_func=_prepare_napari_coords,
        meta_columns=[],
        feature_columns=[],
        valid_extensions=get_valid_extensions(),
    )


def _prepare_napari_coords(
    input_df: pd.DataFrame,
) -> pd.DataFrame:
    output_data: pd.DataFrame = pd.DataFrame(columns=["x", "y", "z"])

    output_data["z"] = input_df["x"]
    output_data["y"] = input_df["y"]
    output_data["x"] = input_df["z"]
    output_data["boxsize"] = DEFAULT_BOXSIZE

    return output_data


def _write_coords(path: os.PathLike, df: pd.DataFrame, **kwargs):
    if "empty_slices" in kwargs:
        print(".coords does not support empty slices.")

    df[["x", "y", "z"]].to_csv(path, sep=" ", header=None, index=None)


def _make_df_data(
    coordinates: pd.DataFrame, box_size: npt.ArrayLike, **kwargs
) -> pd.DataFrame:
    data = {"x": [], "y": [], "z": [], "boxsize": []}

    is_filament_data = coordinates.shape[1] == 4

    for coords, boxsize in zip(
        coordinates,
        box_size,
    ):
        if is_filament_data:
            z, y, x, _ = coords
        else:
            z, y, x = coords

        data["x"].append(x)
        data["y"].append(y)
        data["z"].append(z)
        data["boxsize"].append(boxsize)
    return pd.DataFrame(data)


def from_napari(
    path: os.PathLike,
    layer_data: list[NapariLayerData],
    suffix: str,
    filament_spacing: int,
):
    path = coordsio.from_napari(
        path=path,
        layer_data=layer_data,
        write_func=_write_coords,
        format_func=_make_df_data,
        suffix=suffix,
        filament_spacing=filament_spacing,
    )
    return path



import os
import typing

import mrcfile
import numpy as np
import pandas as pd

from .io_utils import to_napari_image

if typing.TYPE_CHECKING:
    import numpy.typing as npt


def load_image(path: str) -> np.array:

    extension = os.path.splitext(path)[1]
    with mrcfile.mmap(path, "r", permissive=True) as mrc:
        data = mrc.data
        data = data.astype(np.float32)
    if extension == ".mrci":
        return data.astype(np.int32)
    else:
        return data


def get_pixel_size(path: str) -> float:
    with mrcfile.open(path, permissive=True, header_only=True) as mrc:
        return mrc.voxel_size.x if mrc.voxel_size.x != 0 else 1

    return 1

def has_shapes(path: os.PathLike) -> bool:
    return False

def to_napari(
    path: os.PathLike | list[os.PathLike],
) -> "list[tuple[npt.ArrayLike, dict[str, typing.Any], str]]":

    if isinstance(path, list):
        extension = os.path.splitext(path[0])[1]
    else:
        extension = os.path.splitext(path)[1]

    if extension == ".mrci":
        do_normalize = False
    else:
        do_normalize = True

    layer_data = to_napari_image(
        path,
        load_image=load_image,
        get_pixel_size=get_pixel_size,
        do_normalize=do_normalize,
    )

    if extension == ".mrci":
        layer_data = [(_[0], _[1], "labels") for _ in layer_data]

    return layer_data


def get_valid_extensions():
    return ["mrc", "mrcs", "st", "rec", "mrci"]


def from_napari(
    path: os.PathLike | list[os.PathLike] | pd.DataFrame,
    data: typing.Any,
    meta: dict,
):
    raise NotImplementedError



import numpy as np
import pandas as pd
from numpy import testing

from box_manager.io.io_utils import resample_filament


def test_resampling_simple_distance_1():
    input = {"x": [1, 5, 7, 15, 16], "y": [1, 5, 7, 15, 16]}
    expected_output = {
        "x": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16],
        "y": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16],
    }

    fil = pd.DataFrame(input)

    output = resample_filament(fil, 1, ["x", "y"])

    exp_x_float = np.array(expected_output["x"]).astype(float)
    exp_y_float = np.array(expected_output["y"]).astype(float)
    out_x_float = np.array(output["x"]).astype(float)
    out_y_float = np.array(output["y"]).astype(float)
    testing.assert_allclose(exp_x_float, out_x_float)
    testing.assert_allclose(exp_y_float, out_y_float)


def test_resampling_simple_distance_2():
    input = {"x": [1, 5, 7, 15], "y": [1, 5, 7, 15]}
    expected_output = {
        "x": [1, 3, 5, 7, 9, 11, 13, 15],
        "y": [1, 3, 5, 7, 9, 11, 13, 15],
    }

    fil = pd.DataFrame(input)

    output = resample_filament(fil, 2, ["x", "y"])

    exp_x_float = np.array(expected_output["x"]).astype(float)
    exp_y_float = np.array(expected_output["y"]).astype(float)
    out_x_float = np.array(output["x"]).astype(float)
    out_y_float = np.array(output["y"]).astype(float)
    testing.assert_allclose(exp_x_float, out_x_float)
    testing.assert_allclose(exp_y_float, out_y_float)
